//
//  TestDiskController.c
//  RecoveRx
//
//  Created by Sendoh Chen on 2014/7/17.
//
//

#include <libgen.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include "common.h"
#include "filegen.h"
#include "fnctdsk.h"
#include "hdaccess.h"
#include "hdcache.h"
#include "list.h"
#include "partauto.h"
#include "phrecn.h"
#include "photorec.h"
#include "RecoveryUtility.h"
#include "TestDiskController.h"

void (*_cb_status_message)(sRecStatus *recStatus);
void (*_cb_finish_message)(recovery_pstatus_t status);

extern arch_fnct_t arch_none;
bool recovering = false;

void getCurrentTime(char *currentTime)
{
    time_t t = time(NULL);
    struct tm tm = *localtime(&t);
    
    sprintf(currentTime, "%d_%d_%d_%d_%d_%d", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
}

void _postRecovery(const char *path, const char *recup)
{
    char cmd[1024];
    char *dir = basename((char *)recup);
    char *pDir = dirname((char *)recup);
    snprintf(cmd, 1024, "find \"%s\" -maxdepth 1 -name \"%s.*\" -type d -print0 | xargs -0 -J %% mv %% \"%s\"", pDir, dir, path);
    
    //syslog(LOG_NOTICE, cmd);
    runCmd(cmd);
}

void _initRecStatus(sRecStatus *recStatus)
{
    recStatus->progress[0] = '\0';
    recStatus->sectorRead[0] = '\0';
    recStatus->fileFound[0] = '\0';
    recStatus->elapsedTime[0] = '\0';
    recStatus->estimatedTime[0] = '\0';
    for (int i = 0; i < MAX_FILE_TYPE_COUNT; i++) {
        recStatus->fileRecoveredStr[i][0] = '\0';
        recStatus->fileRecoveredNum[i] = 0;
    }
    recStatus->fileRecoveredCount = 0;
}

void _callbackRecoveryFinished(pstatus_t status)
{
    _cb_finish_message((recovery_pstatus_t)status);
    recovering = false;
}

pstatus_t _callbackRecoveryStatus(const struct ph_param *params, const uint64_t offset, const time_t current_time)
{
    if (!recovering)
        return PSTATUS_STOP;
        
    sRecStatus recStatus;
    _initRecStatus(&recStatus);
    
    const partition_t *partition = params->partition;
    const unsigned int sector_size = params->disk->sector_size;
    
    // Progress
    snprintf(recStatus.progress, MAX_BUF_LEN, "%.2f",
             100.0 * ((offset-partition->part_offset)/sector_size) / (partition->part_size/sector_size));
    
    // Reading Sector
    snprintf(recStatus.sectorRead, MAX_BUF_LEN, "%llu/%llu",
             (unsigned long long)((offset-partition->part_offset)/sector_size),
             (unsigned long long)(partition->part_size/sector_size));
    
    // File Found
    if (params->status != STATUS_FIND_OFFSET)
        snprintf(recStatus.fileFound, MAX_BUF_LEN, "%u", params->file_nbr);
    
    if (current_time > params->real_start_time)
    {
        // Elapsed Time
        const time_t elapsed_time = current_time - params->real_start_time;
        snprintf(recStatus.elapsedTime, MAX_BUF_LEN, "%02u:%02u:%02u",
                 (unsigned)(elapsed_time / 3600),
                 (unsigned)((elapsed_time / 60) % 60),
                 (unsigned)(elapsed_time % 60));
        
        // Estimated Time
        if (offset > partition->part_offset && params->status != STATUS_EXT2_ON_BF && params->status != STATUS_EXT2_OFF_BF)
        {
            const time_t eta = (partition->part_offset+partition->part_size-1-offset)*elapsed_time/(offset-partition->part_offset);
            snprintf(recStatus.estimatedTime, MAX_BUF_LEN, "%02u:%02u:%02u",
                     (unsigned)(eta / 3600),
                     (unsigned)((eta / 60) % 60),
                     (unsigned)(eta % 60));
        }
    }

    // File Recovered
    unsigned int i;
    unsigned int nbr;
    file_stat_t *new_file_stats;
    for(i = 0; params->file_stats[i].file_hint != NULL; i++);
    nbr = i;
    if (nbr) {
        new_file_stats = (file_stat_t*)MALLOC(nbr * sizeof(file_stat_t));
        memcpy(new_file_stats, params->file_stats, nbr*sizeof(file_stat_t));
        qsort(new_file_stats, nbr, sizeof(file_stat_t), sorfile_stat_ts);
        for (i = 0; i < nbr && new_file_stats[i].recovered > 0; i++)
        {
            snprintf(recStatus.fileRecoveredStr[i], MAX_BUF_LEN, "%s",
                    (new_file_stats[i].file_hint->extension != NULL ?
                     new_file_stats[i].file_hint->extension : ""));
            
            recStatus.fileRecoveredNum[i] = new_file_stats[i].recovered;
        }

        free(new_file_stats);
    }
    
    // File Recovered Count
    recStatus.fileRecoveredCount = i;
    
    _cb_status_message(&recStatus);
    
    return PSTATUS_OK;
}

void startRecovery(const char *diskName,
                   int partNum,
                   const char *path,
                   char types[][MAX_FILE_TYPE_STR_LEN],
                   int typesCount,
                   void (*cb_status_message)(sRecStatus *recStatus),
                   void (*cb_finish_message)(recovery_pstatus_t status))
{
    _cb_status_message = cb_status_message;
    _cb_finish_message = cb_finish_message;
    
    if (recovering)
        return;
    recovering = true;
    
    int testdisk_mode = TESTDISK_O_RDONLY | TESTDISK_O_READAHEAD_32K;
    list_disk_t *list_disk = NULL;
    list_part_t *list_part = NULL;
    static alloc_data_t list_search_space = {
        .list = TD_LIST_HEAD_INIT(list_search_space.list)
    };
    
    // Init ph_options
    struct ph_options options = {
        .paranoid = 1,
        .keep_corrupted_file = 0,
        .mode_ext2 = 0,
        .expert = 0,
        .lowmem = 0,
        .verbose = 0,
    };
    
    //file_enable_t list_file_enable[typesCount + 1];
    file_enable_t *list_file_enable = (file_enable_t *)malloc(sizeof(file_enable_t) * (typesCount + 1));
    fillFileEnableList(list_file_enable, types, typesCount);
    options.list_file_format = list_file_enable;
    
    // Init ph_param
    struct ph_param params;
    // Add current time in destination path, for dublicate path will move fail
    char currentTime[32];
    getCurrentTime(currentTime);
    params.recup_dir = (char *)malloc(strlen(path) + strlen(currentTime) + 1 + 1);
    strcpy(params.recup_dir, path);
    strcat(params.recup_dir, "_");
    strcat(params.recup_dir, currentTime);
    params.cmd_device = NULL;
    params.cmd_run = NULL;
    params.carve_free_space_only = 0;
    params.blocksize = 0;
    params.recoverx_cb_status = _callbackRecoveryStatus;
    params.recoverx_cb_finished = _callbackRecoveryFinished;
    
    /* random (weak is ok) is need fot GPT */
    //srand(time(NULL));
    
    // Select disk
    list_disk = hd_parse(list_disk, options.verbose, testdisk_mode);
    hd_update_all_geometry(list_disk, options.verbose);
    
    /* Activate the cache, even if photorec has its own */
    for(list_disk_t *element_disk = list_disk;element_disk != NULL; element_disk = element_disk->next)
        element_disk->disk = new_diskcache(element_disk->disk, testdisk_mode);
    
    for (list_disk_t *element_disk = list_disk; element_disk != NULL; element_disk = element_disk->next)
    {
        disk_t *disk = element_disk->disk;
        char *device = disk->device + 5;    // /dev/disk0
        if (strcmp(device, diskName) == 0) {
            params.disk = disk;
            break;
        }
    }
    
    // Arch auto detect
    autodetect_arch(params.disk, &arch_none);
    
    // Select partition
    list_part = init_list_part(params.disk, &options);
    for (int i = 0; i < partNum; i++)
        list_part = list_part->next;
    params.partition = list_part->part;
    
    // Init search space
    if (td_list_empty(&(list_search_space.list)))
        init_search_space(&list_search_space, params.disk, params.partition);
    
    // Do recovery
    photorec(&params, &options, &list_search_space);
    
    // Do post recovery things
    _postRecovery(path, params.recup_dir);
    
    // Free resource
    free(list_file_enable);
    free(params.recup_dir);
    delete_list_disk(list_disk);
    part_free_list(list_part);
    free_list_search_space(&list_search_space);
    free(params.file_stats);
}

void pauseRecovery()
{
    
}

void stopRecovery()
{
    recovering = false;
}